#include <f8/internal/pch.h>
#include <assert.h>

#include <condition_variable>

#include <mysql.h>

#include <a8/mysql.h>
#include <f8/udplog.h>
#include <a8/mutable_xobject.h>

#include <f8/dbpool.h>
#include <f8/msgqueue.h>
#include <f8/utils.h>

namespace f8
{

    enum AsyncQueryError
        {
         AQE_NO_ERROR = 0,
         AQE_EXEC_TYPE_ERROR = 1,
         AQE_QUERY_TYPE_ERROR = 2,
         AQE_SYNTAX_ERROR = 3,
         AQE_CONN_ERROR = 4
        };

    struct AsyncQueryRequest
    {
        long long context_id = 0;
        std::string sql;
#if 1
        std::string _sql_fmt;
        std::vector<a8::XValue> _sql_params;
        a8::XObject conn_info;
#endif
        #if 0
        a8::XParams param;
        #endif
        time_t add_time = 0;
        #if 0
        AsyncDBOnOkFunc on_ok = nullptr;
        AsyncDBOnErrorFunc on_error = nullptr;
        #endif
        f8::Attacher timer_attacher;
    };

    struct AsyncQueryNode
    {
        int socket_handle = 0;
        int query_type = 0;
        long long context_id = 0;
        std::string sql;
#if 1
        std::string _sql_fmt;
        std::vector<a8::XValue> _sql_params;
        a8::XObject conn_info;
#endif
        AsyncQueryNode* nextnode = nullptr;
    };

    class DBThread
    {
    public:

        void Init()
        {
            #if 0
            loop_mutex_ = new std::mutex();
            loop_cond_ = new std::condition_variable();

            last_checkdb_tick_ = a8::XGetTickCount();

            top_node_ = nullptr;
            bot_node_ = nullptr;
            work_node_ = nullptr;
            msg_mutex_ = new std::mutex();
            work_thread_ = new std::thread(&DBThread::WorkThreadProc, this);
            #endif
        }

        void UnInit()
        {
            #if 0
            terminated_ = true;
            work_thread_->join();
            delete work_thread_;
            work_thread_ = nullptr;

            delete msg_mutex_;
            msg_mutex_ = nullptr;

            delete loop_cond_;
            loop_cond_ = nullptr;

            delete loop_mutex_;
            loop_mutex_ = nullptr;
            #endif
        }

        void AddAsyncQuery(AsyncQueryNode* p)
        {
            #if 0
            std::unique_lock<std::mutex> lk(*loop_mutex_);
            msg_mutex_->lock();
            if (bot_node_) {
                bot_node_->nextnode = p;
                bot_node_ = p;
            } else {
                top_node_ = p;
                bot_node_ = p;
            }
            msg_mutex_->unlock();
            loop_cond_->notify_all();
            #endif
        }

    private:

        void WorkThreadProc()
        {
            #if 0
            mysql_thread_init();

            while (!terminated_) {
                CheckDB();
                ProcessMsg();
                ++DBPool::Instance()->run_loop_num;
                WaitLoopCond();
            }

            if (last_query_) {
                delete last_query_;
                last_query_ = nullptr;
            }

            if (last_conn_) {
                delete last_conn_;
                last_conn_ = nullptr;
            }
            mysql_thread_end();
            #endif
        }

        void CheckDB()
        {
            #if 0
            if (a8::XGetTickCount() - last_checkdb_tick_ < 1000 * 60 * 5) {
                return;
            }
            last_checkdb_tick_ = a8::XGetTickCount();
            if (last_conn_ && last_query_) {
                if (last_query_->ExecQuery("SELECT 1;", {}) <= 0) {
#if 0
                    a8::UdpLog::Instance()->Warning("mysql disconnect", {});
                    if (conn.Connect(dbhost_, 3306, dbuser_, dbpasswd_, gamedb_)) {
                        InitMysqlConnection(&query);
                        a8::UdpLog::Instance()->Info("mysql reconnect successed", {});
                    } else {
                        a8::UdpLog::Instance()->Info("mysql reconnect failed", {});
                    }
#endif
                }
            }
            #endif
        }

        void ProcessMsg()
        {
            #if 0
            if (!work_node_ && top_node_) {
                msg_mutex_->lock();
                work_node_ = top_node_;
                top_node_ = nullptr;
                bot_node_ = nullptr;
                msg_mutex_->unlock();
            }
            while (work_node_) {
                AsyncQueryNode *pdelnode = work_node_;
                work_node_ = work_node_->nextnode;
                ProcAsyncQuery(pdelnode);
                ++DBPool::Instance()->exec_query_num;
                delete pdelnode;
            }
            #endif
        }

        void WaitLoopCond()
        {
            #if 0
            std::unique_lock<std::mutex> lk(*loop_mutex_);
            {
                msg_mutex_->lock();
                if (!work_node_ && top_node_) {
                    work_node_ = top_node_;
                    top_node_ = nullptr;
                    bot_node_ = nullptr;
                }
                msg_mutex_->unlock();
            }
            if (!work_node_) {
                loop_cond_->wait_for(lk, std::chrono::seconds(10));
            }
            #endif
        }

        bool ReCreateConn(a8::XObject& conn_info)
        {
            #if 0
            if (last_query_) {
                delete last_query_;
                last_query_ = nullptr;
            }
            if (last_conn_) {
                delete last_conn_;
                last_conn_ = nullptr;
            }
            last_conn_ = new a8::mysql::Connection();
            last_query_ = last_conn_->CreateQuery();
            if (last_conn_->Connect(
                                    conn_info.Get("host"),
                                    conn_info.Get("port"),
                                    conn_info.Get("user"),
                                    conn_info.Get("passwd"),
                                    conn_info.Get("database")
                                    )) {
                f8::InitMysqlConnection(last_query_);
            }
            #endif
            return true;
        }

        bool NeedReCreateConn(a8::XObject& conn_info)
        {
            #if 0
            if (!last_conn_) {
                return true;
            }
            if (last_conn_->GetHost() == conn_info.Get("host").GetString() &&
                last_conn_->GetPort() == conn_info.Get("port").GetInt() &&
                last_conn_->GetUser() == conn_info.Get("user").GetString() &&
                last_conn_->GetPasswd() == conn_info.Get("passwd").GetString() &&
                last_conn_->GetDataBase() == conn_info.Get("database").GetString()
                ) {
                return false;
            }
            #endif
            return true;
        }

        void ProcAsyncQuery(AsyncQueryNode* node)
        {
            #if 0
            if (NeedReCreateConn(node->conn_info)) {
                ReCreateConn(node->conn_info);
            }
            switch (node->query_type) {
            case 0:
                {
                    int ret = last_query_->ExecQuery(node->_sql_fmt.c_str(), node->_sql_params);
                    if (ret < 0) {
                        f8::MsgQueue::Instance()->PostMsg
                            (IM_DbPool,
                             a8::Args
                             (
                              {
                               node->context_id,
                               AQE_SYNTAX_ERROR,
                               last_query_->GetError()
                              }
                              ));
                    } else {
                        DataSet* data_set = new DataSet();
                        data_set->reserve(last_query_->RowsNum());
                        while (!last_query_->Eof()) {
                            auto& row = a8::FastAppend(*data_set);
                            int field_num = last_query_->FieldsNum();
                            row.reserve(field_num);
                            for (int i = 0; i < field_num; i++) {
                                row.push_back(last_query_->GetValue(i).GetString());
                            }
                            last_query_->Next();
                        }
                        f8::MsgQueue::Instance()->PostMsg
                            (IM_DbPool,
                             a8::Args
                             (
                              {
                               node->context_id,
                               AQE_NO_ERROR,
                               data_set
                              }
                              ));
                    }
                }
                break;
            case 1:
                {
                    bool ret = last_query_->ExecScript(node->_sql_fmt.c_str(), node->_sql_params);
                    if (!ret) {
                        MsgQueue::Instance()->PostMsg
                            (IM_DbPool,
                             a8::Args
                             (
                              {
                               node->context_id,
                               AQE_SYNTAX_ERROR,
                               last_query_->GetError()
                              }
                              ));
                    } else {
                        DataSet* data_set = new DataSet();
                        MsgQueue::Instance()->PostMsg
                            (IM_DbPool,
                             a8::Args
                             (
                              {
                               node->context_id,
                               AQE_NO_ERROR,
                               data_set
                              }
                              ));
                    }
                }
                break;
            default:
                {
                    MsgQueue::Instance()->PostMsg
                        (IM_DbPool,
                         a8::Args
                         (
                          {
                           node->context_id,
                           AQE_QUERY_TYPE_ERROR,
                           "wrong query type"
                          }
                          ));
                }
                break;
            }
            #endif
        }

    private:
        volatile bool terminated_ = false;
        std::mutex *loop_mutex_ = nullptr;
        std::condition_variable *loop_cond_ = nullptr;

        a8::mysql::Connection* last_conn_ = nullptr;
        a8::mysql::Query* last_query_ = nullptr;
        long long last_checkdb_tick_ = 0;

        //std::thread *work_thread_ = nullptr;
        AsyncQueryNode *top_node_ = nullptr;
        AsyncQueryNode *bot_node_ = nullptr;
        AsyncQueryNode *work_node_ = nullptr;
        std::mutex *msg_mutex_ = nullptr;
    };

    class DBPoolImpl
    {
    public:

        void Init()
        {
            curr_seqid = 1000001;
        }

        void UnInit()
        {
            for (auto& db_thread : db_thread_pool) {
                db_thread->UnInit();
                delete db_thread;
            }

            for (auto& pair : async_query_hash) {
                delete pair.second;
            }
            async_query_hash.clear();

        }

        void SetThreadNum(int thread_num)
        {
            assert(thread_num > 0);
            for (int i = 0; i < thread_num; i++) {
                DBThread *db_thread = new DBThread();
                db_thread->Init();
                db_thread_pool.push_back(db_thread);
            }
        }

        void AsyncSqlOnOk(long long seqid, DataSet* data_set)
        {
            #if 0
            AsyncQueryRequest* request = GetAsyncQueryRequest(seqid);
            if (!request) {
                return;
            }
            if (request->on_ok) {
                request->on_ok(request->param, data_set);
            }
            async_query_hash.erase(seqid);
            delete request;
            #endif
        }

        void AsyncSqlOnError(long long seqid, int errcode, const std::string& errmsg)
        {
            #if 0
            AsyncQueryRequest* request = GetAsyncQueryRequest(seqid);
            if (!request) {
                return;
            }
            if (DBPool::Instance()->on_dberror) {
                DBPool::Instance()->on_dberror(request->param, errcode, errmsg);
            }
            if (request->on_error) {
                request->on_error(request->param, errcode, errmsg);
            }
            async_query_hash.erase(seqid);
            delete request;
            #endif
        }

        #if 0
        void InternalExecAsyncSql(int exec_type,
                                  a8::XObject& conn_info, const char* querystr, std::vector<a8::XValue>& args,
                                  a8::XParams& param, AsyncDBOnOkFunc on_ok, AsyncDBOnErrorFunc on_error, long long hash_code)
        {
            AsyncQueryRequest* p = new AsyncQueryRequest();
            {
                p->context_id = ++curr_seqid;
                p->param = param;
                p->sql = "";
                p->_sql_fmt = querystr;
                p->_sql_params = args;
                conn_info.DeepCopy(p->conn_info);
                p->add_time = time(nullptr);
                p->on_ok = on_ok;
                p->on_error = on_error;
                async_query_hash[p->context_id] = p;
            }
            DBThread* db_thread = GetDBThread(hash_code);
            if (!db_thread) {
                MsgQueue::Instance()->PostMsg
                    (IM_DbPool,
                     a8::Args
                     (
                      {
                       p->context_id,
                       AQE_CONN_ERROR
                      }
                      ));
                return;
            }
            {
                AsyncQueryNode* node = new AsyncQueryNode();
                node->socket_handle = 0;
                node->query_type = exec_type;
                node->context_id = p->context_id;
                node->sql = "";
                node->_sql_fmt = querystr;
                node->_sql_params = args;
                conn_info.DeepCopy(node->conn_info);
                db_thread->AddAsyncQuery(node);
            }
            f8::Timer::Instance()->SetTimeoutEx
                (1000 * 10,
                 [p] (int event, const a8::Args* args)
                 {
                     if (event == a8::TIMER_EXEC_EVENT) {
                         MsgQueue::Instance()->PostMsg
                             (IM_DbPool,
                              a8::Args
                              (
                               {
                                p->context_id,
                                AQE_CONN_ERROR
                               }
                               ));
                     }
                 },
                 &p->timer_attacher);
            ++DBPool::Instance()->total_query_num;
        }
        #endif

    private:

        AsyncQueryRequest* GetAsyncQueryRequest(long long seqid)
        {
            auto itr = async_query_hash.find(seqid);
            return itr != async_query_hash.end() ? itr->second : nullptr;
        }

        DBThread* GetDBThread(long long hash_code)
        {
            if (db_thread_pool.empty() || hash_code < 0) {
                return nullptr;
            }
            return db_thread_pool[hash_code % db_thread_pool.size()];
        }

    private:
        long long curr_seqid = 0;
        std::map<long long, AsyncQueryRequest*> async_query_hash;

        std::vector<DBThread*> db_thread_pool;
    };

    void DBPool::Init()
    {
        impl_ = std::make_unique<DBPoolImpl>();
        impl_->Init();
        MsgQueue::Instance()->RegisterCallBack
            (IM_DbPool,
             [] (const a8::Args& args)
             {
                 #if 0
                 if (param.param1.GetInt() == AQE_NO_ERROR) {
                     DataSet* data_set = (DataSet*)param.param2.GetUserData();
                     DBPool::Instance()->impl_->AsyncSqlOnOk(param.sender, data_set);
                     delete data_set;
                 } else {
                     DBPool::Instance()->impl_->AsyncSqlOnError(param.sender,
                                                                param.param1,
                                                                param.param2);
                 }
                 #endif
             }
             );
    }

    void DBPool::UnInit()
    {
        impl_->UnInit();
        impl_ = nullptr;
    }

    void DBPool::SetThreadNum(int thread_num)
    {
        impl_->SetThreadNum(thread_num);
    }

    void DBPool::ExecAsyncQuery(int ds,
                                const std::string sql,
                                std::vector<a8::XValue> args,
                                f8::DbQueryResultCb cb)
    {

    }

    void DBPool::ExecAsyncScript(int ds,
                                 const std::string sql,
                                 std::vector<a8::XValue> args,
                                 f8::DbExecResultCb cb)
    {

    }

}
