#pragma once

#include <f8/dataset.h>

namespace f8
{

    class DbOpResult
    {
    public:
        virtual int GetErrCode() = 0;
        virtual std::string GetErrMsg() = 0;
    };

    class DbQueryResult : public DbOpResult
    {
    public:
        virtual std::shared_ptr<f8::DataSet> GetDataSet() = 0;
    };

    class DbExecResult : public DbOpResult
    {
    public:
        virtual long long GetLastInsertId() = 0;
        virtual long long GetRowsAffected() = 0;
    };

    typedef std::function<void(std::shared_ptr<DbQueryResult>)> DbQueryResultCb;
    typedef std::function<void(std::shared_ptr<DbExecResult>)> DbExecResultCb;

    class DBPool
    {
        A8_DECLARE_SINGLETON(DBPool);
    public:
        void Init();
        void UnInit();
        void SetThreadNum(int thread_num);

        //执行异步并行查询
        void ExecAsyncQuery(int ds,
                            const std::string sql,
                            std::vector<a8::XValue> args,
                            f8::DbQueryResultCb cb);
        //执行异步并行sql
        void ExecAsyncScript(int ds,
                             const std::string sql,
                             std::vector<a8::XValue> args,
                             f8::DbExecResultCb cb);

    private:
        std::unique_ptr<class DBPoolImpl> impl_;
    };

}
