#include <f8/internal/pch.h>

#include <f8/tcpclient.h>
#include <f8/internal/asiotcpclient.h>

namespace f8
{

    TcpClient::TcpClient(const std::string& remote_ip, int remote_port)
    {
        impl_ = std::make_shared<TcpClientImpl>(remote_ip, remote_port);
        impl_->on_error =
            [this] (f8::internal::AsioTcpClient*, int err_code)
            {
                if (on_error) {
                    on_error(this, err_code);
                }
            };
        impl_->on_connect =
            [this] (f8::internal::AsioTcpClient*)
            {
                if (on_connect) {
                    on_connect(this);
                }
            };
        impl_->on_disconnect =
            [this] (f8::internal::AsioTcpClient*, int err_code)
            {
                if (on_disconnect) {
                    on_disconnect(this, err_code);
                }
            };
        impl_->on_read =
            [this] (f8::internal::AsioTcpClient*, char* buff, int buff_len)
            {
                if (on_read) {
                    on_read(this, buff, buff_len);
                }
            };
    }

    TcpClient::~TcpClient()
    {
        impl_->Close();
    }

    void TcpClient::Open()
    {
        impl_->Open();
    }

    void TcpClient::Close()
    {
        impl_->Close();
    }

    bool TcpClient::Connected()
    {
        return impl_->Connected();
    }

    void TcpClient::SendBuff(const char* buff, int bufflen)
    {
        impl_->SendBuff(buff, bufflen);
    }

}
