#include <f8/internal/pch.h>

#include <a8/xtimer.h>
#include <f8/timer.h>

namespace f8
{
    class TimerImpl
    {
    public:
        bool initialized_ = false;
        a8::XTimer xtimer_;
    };

    Attacher::Attacher()
    {
        real_obj_.SetOwner(&f8::Timer::Instance()->impl_->xtimer_);
        if (!f8::Timer::Instance()->Initialized()) {
            abort();
        }
    }

    void Attacher::ClearTimerList()
    {
        real_obj_.ClearTimerList();
    }

    Timer::Timer()
    {

    }

    void Timer::Init()
    {
        impl_ = std::make_shared<TimerImpl>();
        impl_->initialized_ = true;
        impl_->xtimer_.Init
            (
             [] (void* context)
             {
                 return a8::XGetTickCount();
             },
             this,
             10,
             1000
             );
    }

    void Timer::UnInit()
    {
        impl_->initialized_ = false;
    }

    bool Timer::Initialized()
    {
        return impl_->initialized_;
    }

    void Timer::Update()
    {
        impl_->xtimer_.Update();
    }

    void Timer::SetTimeout(int expire_time, a8::TimerCb cb)
    {
        impl_->xtimer_.SetTimeout(expire_time, cb);
    }

    void Timer::SetTimeoutEx(int expire_time, a8::TimerCb cb, Attacher* attacher)
    {
        impl_->xtimer_.SetTimeoutEx(expire_time, cb, &attacher->real_obj_);
    }

    TimerWp Timer::SetTimeoutWp(int expire_time, a8::TimerCb cb)
    {
        return TimerWp(impl_->xtimer_.SetTimeoutWp(expire_time, cb));
    }

    TimerWp Timer::SetTimeoutWpEx(int expire_time, a8::TimerCb cb, Attacher* attacher)
    {
        return TimerWp(impl_->xtimer_.SetTimeoutWpEx(expire_time, cb, &attacher->real_obj_));
    }

    void Timer::SetInterval(int expire_time, a8::TimerCb cb)
    {
        impl_->xtimer_.SetInterval(expire_time, cb);
    }

    void Timer::SetIntervalEx(int expire_time, a8::TimerCb cb, Attacher* attacher)
    {
        impl_->xtimer_.SetIntervalEx(expire_time, cb, &attacher->real_obj_);
    }

    TimerWp Timer::SetIntervalWp(int expire_time, a8::TimerCb cb)
    {
        return TimerWp(impl_->xtimer_.SetIntervalWp(expire_time, cb));
    }

    TimerWp Timer::SetIntervalWpEx(int expire_time, a8::TimerCb cb, Attacher* attacher)
    {
        return TimerWp(impl_->xtimer_.SetIntervalWpEx(expire_time, cb, &attacher->real_obj_));
    }

    void Timer::FireEvent(TimerWp& timer_wp, int event, a8::Args* args)
    {
        impl_->xtimer_.FireEvent(timer_wp.real_obj_, event, args);
    }

    void Timer::ModifyTime(TimerWp& timer_wp, int expire_time)
    {
        impl_->xtimer_.ModifyTime(timer_wp.real_obj_, expire_time);
    }

    void Timer::Delete(TimerWp& timer_wp)
    {
        impl_->xtimer_.Delete(timer_wp.real_obj_);
    }

    long long Timer::GetRemainTime(TimerWp& timer_wp)
    {
        return impl_->xtimer_.GetRemainTime(timer_wp.real_obj_);
    }

    void Timer::DeleteCurrentTimer()
    {
        impl_->xtimer_.DeleteCurrentTimer();
    }

    bool Timer::IsRunning()
    {
        return impl_->xtimer_.IsRunning();
    }

    long long Timer::GetIdleTime()
    {
        return impl_->xtimer_.GetIdleTime();
    }

}
